@@ -7,7 +7,7 @@ import typing
 from glob import glob
 import pydicom
 import numpy as np
-from pydicom.tag import Tag
+from pydicom.tag import Tag, BaseTag
 import SimpleITK as sitk
 from skimage.measure import label, regionprops, find_contours
 from threading import Thread
@@ -16,10 +16,12 @@ from queue import *
 import pandas as pd
 import copy
 import cv2
-from typing import List
+from typing import List, Dict
 
 from .Viewer import plot_scroll_Image
 
+PyDicomKeys = Dict[str, BaseTag]
+SitkDicomKeys = Dict[str, str]
 
 def contour_worker(A):
     q, kwargs = A
@@ -63,8 +65,9 @@ def folder_worker(A):
         if item is None:
             break
         else:
-            dicom_adder = AddDicomToDictionary()
-            dicom_path, images_dictionary, rt_dictionary, rd_dictionary, rp_dictionary, verbose = item
+            dicom_path, images_dictionary, rt_dictionary, rd_dictionary, rp_dictionary, verbose, strings = item
+            plan_strings, structure_strings, image_strings, dose_strings = strings
+            dicom_adder = AddDicomToDictionary(plan_strings, structure_strings, image_strings, dose_strings)
             try:
                 if verbose:
                     print('Loading from {}'.format(dicom_path))
@@ -131,10 +134,11 @@ def poly2mask(vertex_row_coords: np.array, vertex_col_coords: np.array,
     coords = np.expand_dims(xy_coords.T, 0)
     mask = np.zeros(shape)
     cv2.fillPoly(mask, coords, 1)
-    return np.array(mask, dtype=np.bool)
+    return np.array(mask, dtype=bool)
 
 
-def add_images_to_dictionary(images_dictionary, sitk_dicom_reader, path: typing.Union[str, bytes, os.PathLike]):
+def add_images_to_dictionary(images_dictionary, dicom_names, sitk_dicom_reader,
+                             path: typing.Union[str, bytes, os.PathLike], sitk_string_keys: SitkDicomKeys = None):
     """
     :param images_dictionary: dictionary of series instance UIDs for images
     :param sitk_dicom_reader: sitk.ImageFileReader()
@@ -155,28 +159,49 @@ def add_images_to_dictionary(images_dictionary, sitk_dicom_reader, path: typing.
         if "0018|0050" in meta_keys:
             slice_thickness = float(sitk_dicom_reader.GetMetaData("0018|0050"))
         study_instance_uid = sitk_dicom_reader.GetMetaData("0020|000d")
-        temp_dict = {'PatientID': patientID, 'SeriesInstanceUID': series_instance_uid,
+        temp_dict = {'PatientID': patientID, 'SeriesInstanceUID': series_instance_uid, 'Files': dicom_names,
                      'StudyInstanceUID': study_instance_uid, 'RTs': {}, 'RDs': {}, 'RPs': {},
                      'Image_Path': path, 'Description': description, 'Pixel_Spacing_X': pixel_spacing_x,
                      'Pixel_Spacing_Y': pixel_spacing_y, 'Slice_Thickness': slice_thickness}
+        if sitk_string_keys is not None:
+            for string in sitk_string_keys:
+                key = sitk_string_keys[string]
+                if key in sitk_dicom_reader.GetMetaDataKeys():
+                    try:
+                        temp_dict[string] = sitk_dicom_reader.GetMetaData(key)
+                    except:
+                        continue
         images_dictionary[series_instance_uid] = temp_dict
 
 
-def add_rp_to_dictionary(ds, path: typing.Union[str, bytes, os.PathLike], rp_dictionary):
+def add_rp_to_dictionary(ds, path: typing.Union[str, bytes, os.PathLike], rp_dictionary,
+                         pydicom_string_keys: PyDicomKeys = None):
     try:
         series_instance_uid = ds.SeriesInstanceUID
         if series_instance_uid not in rp_dictionary:
             refed_structure_uid = ds.ReferencedStructureSetSequence[0].ReferencedSOPInstanceUID
             refed_dose_uid = ds.DoseReferenceSequence[0].DoseReferenceUID
-            temp_dict = {'Path': path, 'SOPInstanceUID': ds.SOPInstanceUID,
+            plan_label = None
+            if Tag((0x300a, 0x002)) in ds.keys():
+                plan_label = ds.RTPlanLabel
+            temp_dict = {'Path': path, 'SOPInstanceUID': ds.SOPInstanceUID, 'PlanLabel': plan_label,
                          'ReferencedStructureSetSOPInstanceUID': refed_structure_uid,
                          'ReferencedDoseSOPUID': refed_dose_uid, 'Description': ds.StudyDescription}
+            if pydicom_string_keys is not None:
+                for string in pydicom_string_keys:
+                    key = pydicom_string_keys[string]
+                    if key in ds.keys():
+                        try:
+                            temp_dict[string] = ds[key].value
+                        except:
+                            continue
             rp_dictionary[series_instance_uid] = temp_dict
     except:
         print("Had an error loading " + path)
 
 
-def add_rt_to_dictionary(ds, path: typing.Union[str, bytes, os.PathLike], rt_dictionary):
+def add_rt_to_dictionary(ds, path: typing.Union[str, bytes, os.PathLike], rt_dictionary,
+                         pydicom_string_keys: PyDicomKeys = None):
     """
     :param ds: pydicom data structure
     :param path: path to the images or structure in question
@@ -202,12 +227,20 @@ def add_rt_to_dictionary(ds, path: typing.Union[str, bytes, os.PathLike], rt_dic
                         temp_dict = {'Path': path, 'ROI_Names': rois, 'ROIs_in_structure': rois_in_structure,
                                      'SeriesInstanceUID': refed_series_instance_uid, 'Plans': {}, 'Doses': {},
                                      'SOPInstanceUID': sop_instance_uid}
+                        if pydicom_string_keys is not None:
+                            for string in pydicom_string_keys:
+                                key = pydicom_string_keys[string]
+                                if key in ds.keys():
+                                    try:
+                                        temp_dict[string] = ds[key].value
+                                    except:
+                                        continue
                         rt_dictionary[series_instance_uid] = temp_dict
     except:
         print("Had an error loading " + path)
 
 
-def add_rd_to_dictionary(sitk_dicom_reader, rd_dictionary):
+def add_rd_to_dictionary(sitk_dicom_reader, rd_dictionary, sitk_string_keys: SitkDicomKeys = None):
     """
     :param ds: pydicom data structure
     :param path: path to the images or structure in question
@@ -215,7 +248,8 @@ def add_rd_to_dictionary(sitk_dicom_reader, rd_dictionary):
     try:
         ds = pydicom.read_file(sitk_dicom_reader.GetFileName())
         series_instance_uid = sitk_dicom_reader.GetMetaData("0020|000e")
-        rt_sopinstance_uid = ds.ReferencedStructureSetSequence[0].ReferencedSOPInstanceUID
+        rt_sopinstance_uid = ds.ReferencedStructureSetSequence[0].ReferencedSOPInstanceUID \
+            if "ReferencedStructureSetSequence" in ds.values() else None
         rp_sopinstance_uid = ds.ReferencedRTPlanSequence[0].ReferencedSOPInstanceUID
         if series_instance_uid not in rd_dictionary:
             study_instance_uid = sitk_dicom_reader.GetMetaData("0020|000d")
@@ -226,6 +260,14 @@ def add_rd_to_dictionary(sitk_dicom_reader, rd_dictionary):
                          'SOPInstanceUID': sitk_dicom_reader.GetMetaData("0008|0018"),
                          'Description': description, 'ReferencedStructureSetSOPInstanceUID': rt_sopinstance_uid,
                          'ReferencedPlanSOPInstanceUID': rp_sopinstance_uid}
+            if sitk_string_keys is not None:
+                for string in sitk_string_keys:
+                    key = sitk_string_keys[string]
+                    if key in sitk_dicom_reader.GetMetaDataKeys():
+                        try:
+                            temp_dict[string] = sitk_dicom_reader.GetMetaData(key)
+                        except:
+                            continue
             rd_dictionary[series_instance_uid] = temp_dict
     except:
         print("Had an error loading " + sitk_dicom_reader.GetFileName())
@@ -251,7 +293,7 @@ def add_sops_to_dictionary(sitk_dicom_reader, series_instances_dictionary):
 
 def return_template_dictionary():
     template_dictionary = {'Image_Path': None, 'PatientID': None, 'RTs': {}, 'RDs': {}, 'RPs': {}, 'Description': None,
-                           'SOP_Instance_UIDs': None, 'SeriesInstanceUID': None}
+                           'SOP_Instance_UIDs': None, 'SeriesInstanceUID': None, 'Files': None}
     return template_dictionary
 
 
@@ -268,11 +310,18 @@ def add_to_mask(mask, z_value, r_value, c_value, mask_value=1):
 
 
 class AddDicomToDictionary(object):
-    def __init__(self):
+    def __init__(self, plan_pydicom_string_keys: PyDicomKeys = Dict or None,
+                 struct_pydicom_string_keys: PyDicomKeys = Dict or None,
+                 image_sitk_string_keys: SitkDicomKeys = Dict or None,
+                 dose_sitk_string_keys: SitkDicomKeys = Dict or None):
         self.image_reader = sitk.ImageFileReader()
         self.image_reader.LoadPrivateTagsOn()
         self.reader = sitk.ImageSeriesReader()
         self.reader.GlobalWarningDisplayOff()
+        self.plan_pydicom_string_keys = plan_pydicom_string_keys
+        self.struct_pydicom_string_keys = struct_pydicom_string_keys
+        self.image_sitk_string_keys = image_sitk_string_keys
+        self.dose_sitk_string_keys = dose_sitk_string_keys
 
     def add_dicom_to_dictionary_from_path(self, dicom_path, images_dictionary, rt_dictionary, rd_dictionary, rp_dictionary):
         fileList = glob(os.path.join(dicom_path, '*.dcm'))
@@ -286,10 +335,11 @@ class AddDicomToDictionary(object):
             modality = self.image_reader.GetMetaData("0008|0060")
             if modality.lower().find('rtdose') != -1:
                 add_rd_to_dictionary(sitk_dicom_reader=self.image_reader,
-                                     rd_dictionary=rd_dictionary)
+                                     rd_dictionary=rd_dictionary, sitk_string_keys=self.dose_sitk_string_keys)
             else:
-                add_images_to_dictionary(images_dictionary=images_dictionary,
-                                         sitk_dicom_reader=self.image_reader, path=dicom_path)
+                add_images_to_dictionary(images_dictionary=images_dictionary, dicom_names=dicom_names,
+                                         sitk_dicom_reader=self.image_reader, path=dicom_path,
+                                         sitk_string_keys=self.image_sitk_string_keys)
         RT_Files = [file for file in fileList if file not in all_names]
         for lstRSFile in RT_Files:
             rt = pydicom.read_file(lstRSFile)
@@ -297,7 +347,8 @@ class AddDicomToDictionary(object):
             if modality.lower().find('struct') != -1:
                 add_rt_to_dictionary(ds=rt, path=lstRSFile, rt_dictionary=rt_dictionary)
             elif modality.lower().find('plan') != -1:
-                add_rp_to_dictionary(ds=rt, path=lstRSFile, rp_dictionary=rp_dictionary)
+                add_rp_to_dictionary(ds=rt, path=lstRSFile, rp_dictionary=rp_dictionary,
+                                     pydicom_string_keys=self.plan_pydicom_string_keys)
         xxx = 1
 
 
@@ -305,7 +356,11 @@ class DicomReaderWriter(object):
     def __init__(self, description='', Contour_Names=None, associations=None, arg_max=True, verbose=True,
                  create_new_RT=True, template_dir=None, delete_previous_rois=True,
                  require_all_contours=True, iteration=0, get_dose_output=False,
-                 flip_axes=(False, False, False), index=0, series_instances_dictionary=None):
+                 flip_axes=(False, False, False), index=0, series_instances_dictionary=None,
+                 plan_pydicom_string_keys: PyDicomKeys = None,
+                 struct_pydicom_string_keys: PyDicomKeys = None,
+                 image_sitk_string_keys: SitkDicomKeys = None,
+                 dose_sitk_string_keys: SitkDicomKeys = None):
         """
         :param description: string, description information to add to .nii files
         :param delete_previous_rois: delete the previous RTs within the structure when writing out a prediction
@@ -322,6 +377,10 @@ class DicomReaderWriter(object):
         :param series_instances_dictionary: dictionary of series instance UIDs of images and RTs
         """
         self.rt_dictionary = {}
+        self.plan_pydicom_string_keys = plan_pydicom_string_keys
+        self.struct_pydicom_string_keys = struct_pydicom_string_keys
+        self.image_sitk_string_keys = image_sitk_string_keys
+        self.dose_sitk_string_keys = dose_sitk_string_keys
         self.images_dictionary = {}
         self.rd_dictionary = {}
         self.rp_dictionary = {}
@@ -350,11 +409,14 @@ class DicomReaderWriter(object):
         self.template = True
         self.delete_previous_rois = delete_previous_rois
         self.reader = sitk.ImageSeriesReader()
+        self.image_reader = sitk.ImageFileReader()
+        self.image_reader.LoadPrivateTagsOn()
         self.reader.MetaDataDictionaryArrayUpdateOn()
         self.reader.LoadPrivateTagsOn()
         self.reader.SetOutputPixelType(sitk.sitkFloat32)
         self.verbose = verbose
         self.dicom_handle_uid = None
+        self.dicom_info_uid = None
         self.RS_struct_uid = None
         self.rd_study_instance_uid = None
         self.index = index
@@ -403,6 +465,105 @@ class DicomReaderWriter(object):
                 template = return_template_dictionary()
                 template['RTs'].update({rt_series_instance_uid: self.rt_dictionary[rt_series_instance_uid]})
                 self.series_instances_dictionary[index] = template
+        for rd_series_instance_uid in self.rd_dictionary:
+            struct_ref = self.rd_dictionary[rd_series_instance_uid]['ReferencedStructureSetSOPInstanceUID']
+            if struct_ref is None:
+                continue
+            for image_series_key in self.series_instances_dictionary:
+                rts = self.series_instances_dictionary[image_series_key]['RTs']
+                for rt_key in rts:
+                    structure_sop_uid = rts[rt_key]['SOPInstanceUID']
+                    if struct_ref == structure_sop_uid:
+                        rts[rt_key]['Doses'][rd_series_instance_uid] = self.rd_dictionary[rd_series_instance_uid]
+                        self.series_instances_dictionary[image_series_key]['RDs'].update({rd_series_instance_uid:
+                                                                                              self.rd_dictionary[rd_series_instance_uid]})
+        for rp_series_instance_uid in self.rp_dictionary:
+            added = False
+            struct_ref = self.rp_dictionary[rp_series_instance_uid]['ReferencedStructureSetSOPInstanceUID']
+            for image_series_key in self.series_instances_dictionary:
+                rts = self.series_instances_dictionary[image_series_key]['RTs']
+                for rt_key in rts:
+                    structure_sop_uid = rts[rt_key]['SOPInstanceUID']
+                    if struct_ref == structure_sop_uid:
+                        rts[rt_key]['Plans'][rp_series_instance_uid] = self.rp_dictionary[rp_series_instance_uid]
+                        self.series_instances_dictionary[image_series_key]['RPs'].update({rp_series_instance_uid:
+                                                                                              self.rp_dictionary[rp_series_instance_uid]})
+                        added = True
+            if not added:
+                while index in self.series_instances_dictionary:
+                    index += 1
+                template = return_template_dictionary()
+                template['RPs'].update({rp_series_instance_uid: self.rp_dictionary[rp_series_instance_uid]})
+                self.series_instances_dictionary[index] = template
+        for rd_series_instance_uid in self.rd_dictionary:
+            added = False
+            struct_ref = self.rd_dictionary[rd_series_instance_uid]['ReferencedStructureSetSOPInstanceUID']
+            if struct_ref is not None:
+                continue
+            plan_ref = self.rd_dictionary[rd_series_instance_uid]['ReferencedPlanSOPInstanceUID']
+            for image_series_key in self.series_instances_dictionary:
+                rps = self.series_instances_dictionary[image_series_key]['RPs']
+                rts = self.series_instances_dictionary[image_series_key]['RTs']
+                for rp_key in rps:
+                    plan_sop_uid = rps[rp_key]['SOPInstanceUID']
+                    if plan_ref == plan_sop_uid:
+                        rt_key_sopinstanceUID = rps[rp_key]['ReferencedStructureSetSOPInstanceUID']
+                        for rt_key in rts:
+                            if rts[rt_key]['SOPInstanceUID'] == rt_key_sopinstanceUID:
+                                rts[rt_key]['Doses'][rd_series_instance_uid] = self.rd_dictionary[rd_series_instance_uid]
+                        self.series_instances_dictionary[image_series_key]['RDs'].update({rd_series_instance_uid:
+                                                                                              self.rd_dictionary[rd_series_instance_uid]})
+                        added = True
+            if not added:
+                while index in self.series_instances_dictionary:
+                    index += 1
+                template = return_template_dictionary()
+                template['RDs'].update({rd_series_instance_uid: self.rd_dictionary[rd_series_instance_uid]})
+                self.series_instances_dictionary[index] = template
+
+    def __manual_compile_based_on_folders__(self, reset_series_instances_dict=False):
+        """
+        The goal of this is to combine image, rt, and dose dictionaries based on folder location
+        AKA, if the RT structure and images are in the same folder
+        :return:
+        """
+        print("Don't use this unless you know why you're doing it...")
+        if reset_series_instances_dict:
+            self.series_instances_dictionary = {}
+        if self.verbose:
+            print('Compiling dictionaries together...')
+        folders = []
+        for key, value in self.series_instances_dictionary.items():
+            folders.append(value['Image_Path'])
+        index = 0
+        image_keys = list(self.images_dictionary.keys())
+        image_keys.sort()
+        for series_instance_uid in image_keys:  # Will help keep things in order later
+            folder = self.images_dictionary[series_instance_uid]['Image_Path']
+            if folder not in folders:
+                while index in self.series_instances_dictionary:
+                    index += 1
+                self.series_instances_dictionary[index] = self.images_dictionary[series_instance_uid]
+                folders.append(folder)
+        for rt_series_instance_uid in self.rt_dictionary:
+            rt_path = os.path.split(self.rt_dictionary[rt_series_instance_uid]['Path'])[0]
+            rt_dictionary = self.rt_dictionary[rt_series_instance_uid]
+            path = rt_dictionary['Path']
+            self.all_RTs[path] = rt_dictionary['ROI_Names']
+            for roi in rt_dictionary['ROI_Names']:
+                if roi not in self.RTs_with_ROI_Names:
+                    self.RTs_with_ROI_Names[roi] = [path]
+                else:
+                    self.RTs_with_ROI_Names[roi].append(path)
+            if rt_path in folders:
+                index = folders.index(rt_path)
+                self.series_instances_dictionary[index]['RTs'].update({rt_series_instance_uid: self.rt_dictionary[rt_series_instance_uid]})
+            else:
+                while index in self.series_instances_dictionary:
+                    index += 1
+                template = return_template_dictionary()
+                template['RTs'].update({rt_series_instance_uid: self.rt_dictionary[rt_series_instance_uid]})
+                self.series_instances_dictionary[index] = template
         for rd_series_instance_uid in self.rd_dictionary:
             added = False
             struct_ref = self.rd_dictionary[rd_series_instance_uid]['ReferencedStructureSetSOPInstanceUID']
@@ -439,11 +600,13 @@ class DicomReaderWriter(object):
                 template = return_template_dictionary()
                 template['RPs'].update({rp_series_instance_uid: self.rp_dictionary[rp_series_instance_uid]})
                 self.series_instances_dictionary[index] = template
+        self.__check_if_all_contours_present__()
 
     def __reset__(self):
         self.__reset_RTs__()
         self.rd_study_instance_uid = None
         self.dicom_handle_uid = None
+        self.dicom_info_uid = None
         self.series_instances_dictionary = {}
         self.rt_dictionary = {}
         self.images_dictionary = {}
@@ -665,7 +828,8 @@ class DicomReaderWriter(object):
                 threads.append(t)
             for index, path in enumerate(paths_with_dicom):
                 item = [path, self.images_dictionary, self.rt_dictionary, self.rd_dictionary, self.rp_dictionary,
-                        self.verbose]
+                        self.verbose, (self.plan_pydicom_string_keys, self.struct_pydicom_string_keys,
+                                       self.image_sitk_string_keys, self.dose_sitk_string_keys)]
                 q.put(item)
             for i in range(thread_count):
                 q.put(None)
@@ -793,25 +957,62 @@ class DicomReaderWriter(object):
             df.to_excel(excel_file, index=0)
 
     def get_images_and_mask(self) -> None:
-        assert self.index in self.series_instances_dictionary, \
-            'Index is not present in the dictionary! Set it using set_index(index)'
+        if self.index not in self.series_instances_dictionary:
+            print("Index is not preset in the dictionary! Set it using set_index(index)")
+            return None
         self.get_images()
         self.get_mask()
         if self.get_dose_output:
             self.get_dose()
 
-    def get_images(self) -> None:
-        assert self.index in self.series_instances_dictionary, \
-            'Index is not present in the dictionary! Set it using set_index(index)'
+    def get_all_info(self) -> None:
+        """
+        Print all of the keys and their respective values
+        :return:
+        """
+        self.load_key_information_only()
+        for key in self.image_reader.GetMetaDataKeys():
+            print("{} is {}".format(key, self.image_reader.GetMetaData(key)))
+
+    def return_key_info(self, key):
+        """
+        Return the dicom information for a particular key
+        Example: "0008|0022" will return the date acquired in YYYYMMDD format
+        :param key: dicom key "0008|0022"
+        :return: value associated with the key
+        """
+        self.load_key_information_only()
+        if not self.image_reader.HasMetaDataKey(key):
+            print("{} is not present in the reader".format(key))
+            return None
+        return self.image_reader.GetMetaData(key)
+
+    def load_key_information_only(self) -> None:
+        if self.index not in self.series_instances_dictionary:
+            print('Index is not present in the dictionary! Set it using set_index(index)')
+            return None
         index = self.index
         series_instance_uid = self.series_instances_dictionary[index]['SeriesInstanceUID']
+        if self.dicom_info_uid != series_instance_uid:  # Only load if needed
+            dicom_names = self.series_instances_dictionary[index]['Files']
+            self.image_reader.SetFileName(dicom_names[0])
+            self.image_reader.ReadImageInformation()
+            self.dicom_info_uid = series_instance_uid
+
+    def get_images(self) -> None:
+        if self.index not in self.series_instances_dictionary:
+            print('Index is not present in the dictionary! Set it using set_index(index)')
+            return None
+        index = self.index
+        series_instance_uid = self.series_instances_dictionary[index]['SeriesInstanceUID']
+        if series_instance_uid is None:
+            print("This index does not have an associated image within the loaded folders")
+            return None
         if self.dicom_handle_uid != series_instance_uid:  # Only load if needed
             if self.verbose:
                 print('Loading images for {} at \n {}\n'.format(self.series_instances_dictionary[index]['Description'],
                                                                 self.series_instances_dictionary[index]['Image_Path']))
-
-            dicom_path = self.series_instances_dictionary[index]['Image_Path']
-            dicom_names = self.reader.GetGDCMSeriesFileNames(dicom_path, series_instance_uid)
+            dicom_names = self.series_instances_dictionary[index]['Files']
             self.ds = pydicom.read_file(dicom_names[0])
             self.reader.SetFileNames(dicom_names)
             self.dicom_handle = self.reader.Execute()
@@ -826,8 +1027,9 @@ class DicomReaderWriter(object):
             self.dicom_handle_uid = series_instance_uid
 
     def get_dose(self) -> None:
-        assert self.index in self.series_instances_dictionary, \
-            'Index is not present in the dictionary! Set it using set_index(index)'
+        if self.index not in self.series_instances_dictionary:
+            print('Index is not present in the dictionary! Set it using set_index(index)')
+            return None
         index = self.index
         if self.dicom_handle_uid != self.series_instances_dictionary[index]['SeriesInstanceUID']:
             print('Loading images for index {}, since mask was requested but image loading was '
@@ -864,12 +1066,20 @@ class DicomReaderWriter(object):
             output.SetOrigin(origin)
             self.dose_handle = output
 
+    def __mask_empty_mask__(self) -> None:
+        self.mask = np.zeros(
+            [self.dicom_handle.GetSize()[-1], self.image_size_rows, self.image_size_cols, len(self.Contour_Names) + 1],
+            dtype='int8')
+
     def get_mask(self) -> None:
-        assert self.index in self.series_instances_dictionary, \
-            'Index is not present in the dictionary! Set it using set_index(index)'
-        assert self.Contour_Names, 'If you want a mask, you need to set the contour names you are looking ' \
-                                   'for, use set_contour_names_and_associations(list_of_roi_names).\nIf you just' \
-                                   ' want to look at images  use get_images() not get_images_and_mask() or get_mask()'
+        if self.index not in self.series_instances_dictionary:
+            print('Index is not present in the dictionary! Set it using set_index(index)')
+            return None
+        if not self.Contour_Names:
+            print('If you want a mask, you need to set the contour names you are looking for, use '
+                  'set_contour_names_and_associations(list_of_roi_names).\nIf you just '
+                  'want to look at images  use get_images() not get_images_and_mask() or get_mask()')
+            return None
         index = self.index
         if self.dicom_handle_uid != self.series_instances_dictionary[index]['SeriesInstanceUID']:
             print('Loading images for index {}, since mask was requested but image loading was '
@@ -877,9 +1087,7 @@ class DicomReaderWriter(object):
             self.get_images()
         if self.RS_struct_uid == self.series_instances_dictionary[index]['SeriesInstanceUID']:  # Already loaded
             return None
-        self.mask = np.zeros(
-            [self.dicom_handle.GetSize()[-1], self.image_size_rows, self.image_size_cols, len(self.Contour_Names) + 1],
-            dtype='int8')
+        self.__mask_empty_mask__()
         RTs = self.series_instances_dictionary[index]['RTs']
         for RT_key in RTs:
             RT = RTs[RT_key]
@@ -920,61 +1128,77 @@ class DicomReaderWriter(object):
         self.annotation_handle.SetDirection(self.dicom_handle.GetDirection())
         return None
 
+    def reshape_contour_data(self, as_array: np.array):
+        as_array = np.asarray(as_array)
+        if as_array.shape[-1] != 3:
+            as_array = np.reshape(as_array, [as_array.shape[0] // 3, 3])
+        matrix_points = np.asarray([self.dicom_handle.TransformPhysicalPointToIndex(as_array[i])
+                                    for i in range(as_array.shape[0])])
+        return matrix_points
+
+    def return_mask(self, mask: np.array, matrix_points: np.array, geometric_type: str):
+        col_val = matrix_points[:, 0]
+        row_val = matrix_points[:, 1]
+        z_vals = matrix_points[:, 2]
+        if geometric_type != "OPEN_NONPLANAR":
+            temp_mask = poly2mask(row_val, col_val, [self.image_size_rows, self.image_size_cols])
+            # temp_mask[self.row_val, self.col_val] = 0
+            mask[z_vals[0], temp_mask] += 1
+        else:
+            for point_index in range(len(z_vals) - 1, 0, -1):
+                z_start = z_vals[point_index]
+                z_stop = z_vals[point_index - 1]
+                z_dif = z_stop - z_start
+                r_start = row_val[point_index]
+                r_stop = row_val[point_index - 1]
+                r_dif = r_stop - r_start
+                c_start = col_val[point_index]
+                c_stop = col_val[point_index - 1]
+                c_dif = c_stop - c_start
+
+                step = 1
+                if z_dif != 0:
+                    r_slope = r_dif / z_dif
+                    c_slope = c_dif / z_dif
+                    if z_dif < 0:
+                        step = -1
+                    for z_value in range(z_start, z_stop + step, step):
+                        r_value = r_start + r_slope * (z_value - z_start)
+                        c_value = c_start + c_slope * (z_value - z_start)
+                        add_to_mask(mask=mask, z_value=z_value, r_value=r_value, c_value=c_value)
+                if r_dif != 0:
+                    c_slope = c_dif / r_dif
+                    z_slope = z_dif / r_dif
+                    if r_dif < 0:
+                        step = -1
+                    for r_value in range(r_start, r_stop + step, step):
+                        c_value = c_start + c_slope * (r_value - r_start)
+                        z_value = z_start + z_slope * (r_value - r_start)
+                        add_to_mask(mask=mask, z_value=z_value, r_value=r_value, c_value=c_value)
+                if c_dif != 0:
+                    r_slope = r_dif / c_dif
+                    z_slope = z_dif / c_dif
+                    if c_dif < 0:
+                        step = -1
+                    for c_value in range(c_start, c_stop + step, step):
+                        r_value = r_start + r_slope * (c_value - c_start)
+                        z_value = z_start + z_slope * (c_value - c_start)
+                        add_to_mask(mask=mask, z_value=z_value, r_value=r_value, c_value=c_value)
+        return mask
+
+    def contour_points_to_mask(self, contour_points, mask=None):
+        if mask is None:
+            mask = np.zeros([self.dicom_handle.GetSize()[-1], self.image_size_rows, self.image_size_cols], dtype='int8')
+        matrix_points = self.reshape_contour_data(contour_points)
+        mask = self.return_mask(mask, matrix_points, geometric_type="CLOSED_PLANAR")
+        return mask
+
     def contours_to_mask(self, index: int):
         mask = np.zeros([self.dicom_handle.GetSize()[-1], self.image_size_rows, self.image_size_cols], dtype='int8')
         Contour_data = self.RS_struct.ROIContourSequence[index].ContourSequence
         for i in range(len(Contour_data)):
-            as_array = np.asarray(Contour_data[i].ContourData[:])
-            reshaped = np.reshape(as_array, [as_array.shape[0] // 3, 3])
-            matrix_points = np.asarray([self.dicom_handle.TransformPhysicalPointToIndex(reshaped[i])
-                                        for i in range(reshaped.shape[0])])
-            self.col_val = matrix_points[:, 0]
-            self.row_val = matrix_points[:, 1]
-            z_vals = matrix_points[:, 2]
-            if Contour_data[i].ContourGeometricType != 'OPEN_NONPLANAR':
-                temp_mask = poly2mask(self.row_val, self.col_val, [self.image_size_rows, self.image_size_cols])
-                # temp_mask[self.row_val, self.col_val] = 0
-                mask[z_vals[0], temp_mask] += 1
-            else:
-                for point_index in range(len(z_vals)-1, 0, -1):
-                    z_start = z_vals[point_index]
-                    z_stop = z_vals[point_index - 1]
-                    z_dif = z_stop - z_start
-                    r_start = self.row_val[point_index]
-                    r_stop = self.row_val[point_index - 1]
-                    r_dif = r_stop - r_start
-                    c_start = self.col_val[point_index]
-                    c_stop = self.col_val[point_index - 1]
-                    c_dif = c_stop - c_start
-
-                    step = 1
-                    if z_dif != 0:
-                        r_slope = r_dif / z_dif
-                        c_slope = c_dif / z_dif
-                        if z_dif < 0:
-                            step = -1
-                        for z_value in range(z_start, z_stop + step, step):
-                            r_value = r_start + r_slope * (z_value - z_start)
-                            c_value = c_start + c_slope * (z_value - z_start)
-                            add_to_mask(mask=mask, z_value=z_value, r_value=r_value, c_value=c_value)
-                    if r_dif != 0:
-                        c_slope = c_dif / r_dif
-                        z_slope = z_dif / r_dif
-                        if r_dif < 0:
-                            step = -1
-                        for r_value in range(r_start, r_stop + step, step):
-                            c_value = c_start + c_slope * (r_value - r_start)
-                            z_value = z_start + z_slope * (r_value - r_start)
-                            add_to_mask(mask=mask, z_value=z_value, r_value=r_value, c_value=c_value)
-                    if c_dif != 0:
-                        r_slope = r_dif / c_dif
-                        z_slope = z_dif / c_dif
-                        if c_dif < 0:
-                            step = -1
-                        for c_value in range(c_start, c_stop + step, step):
-                            r_value = r_start + r_slope * (c_value - c_start)
-                            z_value = z_start + z_slope * (c_value - c_start)
-                            add_to_mask(mask=mask, z_value=z_value, r_value=r_value, c_value=c_value)
+            matrix_points = self.reshape_contour_data(Contour_data[i].ContourData[:])
+            mask = self.return_mask(mask, matrix_points, geometric_type=Contour_data[i].ContourGeometricType)
         mask = mask % 2
         return mask
 
@@ -1013,19 +1237,23 @@ class DicomReaderWriter(object):
         fid.close()
 
     def prediction_array_to_RT(self, prediction_array: np.array, output_dir: typing.Union[str, bytes, os.PathLike],
-                               ROI_Names: List[str]):
+                               ROI_Names: List[str]) -> None:
         """
         :param prediction_array: numpy array of prediction, expected shape is [#Images, Rows, Cols, #Classes + 1]
         :param output_dir: directory to pass RT structure to
         :param ROI_Names: list of ROI names equal to the number of classes
         :return:
         """
-        assert ROI_Names is not None, 'You need to provide ROI_Names'
-        assert prediction_array.shape[-1] == len(ROI_Names) + 1, 'Your last dimension of prediction array should be' \
-                                                                 ' equal  to the number or ROI_names minus 1, channel' \
-                                                                 ' 0 is background'
-        assert self.index in self.series_instances_dictionary, \
-            'Index is not present in the dictionary! Set it using set_index(index)'
+        if ROI_Names is None:
+            print("You need to provide ROI_Names")
+            return None
+        if prediction_array.shape[-1] != (len(ROI_Names) + 1):
+            print("Your last dimension of prediction array should be equal  to the number or ROI_names minus 1,"
+                  "channel. 0 is for background")
+            return None
+        if self.index not in self.series_instances_dictionary:
+            print("Index is not present in the dictionary! Set it using set_index(index)")
+            return None
         index = self.index
         if self.dicom_handle_uid != self.series_instances_dictionary[index]['SeriesInstanceUID']:
             self.get_images()
